package edu.sc.seis.cormorant.event;

import java.sql.SQLException;
import java.util.Properties;

import org.omg.CORBA.ORB;
import org.omg.CORBA.ORBPackage.InvalidName;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.Servant;
import org.omg.PortableServer.ServantManager;
import org.omg.PortableServer.POAPackage.AdapterAlreadyExists;
import org.omg.PortableServer.POAPackage.InvalidPolicy;
import org.omg.PortableServer.POAPackage.WrongPolicy;

import edu.iris.Fissures.IfEvent.EventFinder;
import edu.sc.seis.cormorant.DBProvidingAbstractController;
import edu.sc.seis.fissuresUtil.hibernate.EventDB;
import edu.sc.seis.fissuresUtil.hibernate.HibernateUtil;

public class EventController extends DBProvidingAbstractController {

    public EventController(Properties confProps, String serverPropName, ORB orb)
            throws Exception {
        super(confProps, serverPropName, orb);
        createEventDC(orb, confProps);
    }

    public Servant getServant() {
        return eventDC;
    }

    protected EventFinderImpl getEventFinderImpl(POA eventControllerPOA, EventDataAccess eventDao)
            throws SQLException {
        return new EventFinderImpl(eventDao,
                                   eventControllerPOA);
    }
    
    protected EventDataAccess createEventDataAccess(ORB orb, Properties props) {
        synchronized(HibernateUtil.class) {
            HibernateUtil.setUpFromConnMgr(props, HibernateUtil.DEFAULT_EHCACHE_CONFIG);
        }
        return new EventDBDataAccess(EventDB.getSingleton());
    }

    protected void createEventDC(ORB orb, Properties props) throws SQLException,
            AdapterAlreadyExists, InvalidPolicy, InvalidName, WrongPolicy {
        EventDataAccess jdbcEventAccess = createEventDataAccess(orb, props);
        POA eventControllerPOA = createPOA(orb);
        eventDC = new EventDCImpl();
        EventFinderImpl finderImpl = getEventFinderImpl(eventControllerPOA, jdbcEventAccess);
        EventFinder finder = finderImpl._this(orb);
        finderImpl.setEventFinder(finder);
        eventDC.setEventFinder(finder);
        EventLocator_impl eimpl = new EventLocator_impl(jdbcEventAccess,
                                                        eventControllerPOA);
        ServantManager locator = eimpl._this(orb);
        eventControllerPOA.set_servant_manager(locator);
    }

    private EventDCImpl eventDC;
}