package edu.sc.seis.cormorant.nameService;

/*
 * JacORB - a free Java ORB Copyright (C) 1997-2003 Gerald Brose. This library
 * is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Library General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version. This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
 * for more details. You should have received a copy of the GNU Library General
 * Public License along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import java.util.HashMap;
import java.util.Properties;

import org.jacorb.naming.Name;
import org.omg.CORBA.NO_IMPLEMENT;
import org.omg.CORBA.ORB;
import org.omg.CORBA.Object;
import org.omg.CosNaming.BindingIteratorHolder;
import org.omg.CosNaming.BindingListHolder;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextExtPOA;
import org.omg.CosNaming.NamingContextPackage.AlreadyBound;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.InvalidName;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.CosNaming.NamingContextPackage.NotFoundReason;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAPackage.ObjectNotActive;
import org.omg.PortableServer.POAPackage.ServantAlreadyActive;
import org.omg.PortableServer.POAPackage.WrongPolicy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The actual implementation for the CORBAService Naming
 * 
 * @author Gerald Brose, FU Berlin
 * @version $Id: NamingContextImpl.java,v 1.22 2004/02/05 15:57:12 nick.cross
 *          Exp $
 */
public class NamingContextImpl extends NamingContextExtPOA {

    public NamingContextImpl(Properties confProps,
                             String serverPropName,
                             ORB orb,
                             POA poa) {
        this.poa = poa;
        this.orb = orb;
        this.serverPropsName = serverPropName;
        this.confProps = confProps;
        this.subcontext = new HashMap();
        this.corbaObject = new HashMap();
    }

    public String to_string(NameComponent[] n) {
        throw new NO_IMPLEMENT();
    }

    public NameComponent[] to_name(String sn) throws InvalidName {
        return Name.toName(sn);
    }

    public String to_url(String addr, String sn) {
        throw new NO_IMPLEMENT();
    }

    public Object resolve_str(String sn) {
        throw new NO_IMPLEMENT();
    }

    public void bind(NameComponent[] name, Object obj) throws NotFound,
            CannotProceed, AlreadyBound, InvalidName {
        try {
            // Does not currently throw InvalidName exception
            if(name.length == 0) {
                logger.error("Empty NameComponent array.");
                throw new CannotProceed();
            } else if(name.length == 1) {
                logger.debug("Attempting to bind " + obj.toString() + " to "
                        + this.toString(name[0]));
                if(!this.containsName(name[0])) {
                    this.corbaObject.put(this.toString(name[0]), obj);
                } else {
                    throw new AlreadyBound();
                }
            } else {
                if(this.containsName(name[0])) {
                    NameComponent[] croppedName = this.cropName(name);
                    this.getSubcontext(name).bind(croppedName, obj);
                } else {
                    logger.debug("Name server does not contain "
                            + this.toString(name[0]));
                    NameComponent[] namesMinusObject = this.cropObject(name);
                    throw new NotFound(NotFoundReason.missing_node,
                                       namesMinusObject);
                }
            }
        } catch(NotFound e) {
            logger.debug("Context not found. Try creating it first.");
            throw e;
        }
    }

    private NamingContextExt getSubcontext(NameComponent[] name) {
        return ((NamingContextExt)this.subcontext.get(toString(name[0])));
    }

    private String toString(NameComponent name) {
        return name.id + "." + name.kind;
    }

    public void bind_context(NameComponent[] n, NamingContext nc) {
        throw new NO_IMPLEMENT();
    }

    public void rebind(NameComponent[] name, Object obj) throws NotFound,
            CannotProceed, InvalidName {
        try {
            if(name.length == 0) {
                logger.error("Empty NameComponent array.");
                throw new CannotProceed();
            } else if(name.length == 1) {
                logger.debug("Rebinding " + obj.toString() + " to "
                        + this.toString(name[0]));
                this.corbaObject.put(this.toString(name[0]), obj);
            } else {
                if(this.containsName(name[0])) {
                    NameComponent[] croppedName = this.cropName(name);
                    this.getSubcontext(name).rebind(croppedName, obj);
                } else {
                    logger.debug("Name server does not contain "
                            + this.toString(name[0]));
                    NameComponent[] namesMinusObject = this.cropObject(name);
                    throw new NotFound(NotFoundReason.missing_node,
                                       namesMinusObject);
                }
            }
        } catch(NotFound e) {
            logger.debug("Context not found. Try creating it first.");
            throw e;
        }
    }

    public void rebind_context(NameComponent[] n, NamingContext nc) {
        throw new NO_IMPLEMENT();
    }

    public Object resolve(NameComponent[] name) throws NotFound, CannotProceed,
            InvalidName {
        try {
            if(name.length == 0) {
                logger.error("Empty NameComponent array.");
                throw new CannotProceed();
            } else if(name.length == 1) {
                logger.debug("Attempting to resolve " + this.toString(name[0]));
                return (Object)this.corbaObject.get(this.toString(name[0]));
            } else {   
                if(this.containsName(name[0])) {
                    NameComponent[] croppedName = this.cropName(name);
                    return this.getSubcontext(name).resolve(croppedName);
                } else {
                    logger.debug("Name server does not contain "
                            + this.toString(name[0]));
                    NameComponent[] namesMinusObject = this.cropObject(name);
                    throw new NotFound(NotFoundReason.missing_node,
                                       namesMinusObject);
                }
            }
        } catch(NotFound e) {
            logger.debug("Context not found. Try creating it first.");
            throw e;
        }
    }

    public void unbind(NameComponent[] n) {
        throw new NO_IMPLEMENT();
    }

    public void list(int how_many,
                     BindingListHolder bl,
                     BindingIteratorHolder bi) {
        throw new NO_IMPLEMENT();
    }

    public NamingContext new_context() {
        throw new NO_IMPLEMENT();
    }

    public NamingContext bind_new_context(NameComponent[] name)
            throws NotFound, AlreadyBound, CannotProceed, InvalidName {
        try {
            if(name.length == 0) {
                logger.error("Empty NameComponent array.");
                throw new CannotProceed();
            } else if(name.length == 1) {
                logger.debug("Attempting to bind_new_context "
                        + this.toString(name[0]));
                NamingContextImpl namingContextImpl = new NamingContextImpl(this.confProps,
                                                                            this.serverPropsName,
                                                                            this.orb,
                                                                            this.poa);
                byte[] id;
                Object namingContextReference;
                // Check if name already bound
                if(!containsName(name[0])) {
                    try {
                        id = this.poa.activate_object(namingContextImpl);
                        namingContextReference = this.poa.id_to_reference(id);
                    } catch(ServantAlreadyActive e) {
                        logger.error("Servant already active.");
                        throw new RuntimeException("I checked for it being bound above, but that must've failed.",
                                                   e);
                    } catch(WrongPolicy e) {
                        logger.error("Wrong policy.");
                        throw new RuntimeException("The policy should have been checked to be valid, but that must've failed.",
                                                   e);
                    } catch(ObjectNotActive e) {
                        logger.error("Object not active.");
                        throw new RuntimeException("The object should have been checked to be active, but that must've failed.",
                                                   e);
                    }
                    NamingContextExt namingContextExt = NamingContextExtHelper.narrow(namingContextReference);
                    this.subcontext.put(this.toString(name[0]),
                                        namingContextExt);
                    return namingContextExt;
                } else {
                    throw new AlreadyBound();
                }
            } else {
                if(this.containsName(name[0])) {
                    NameComponent[] croppedName = this.cropName(name);
                    return this.getSubcontext(name)
                            .bind_new_context(croppedName);
                } else {
                    logger.debug("Name server does not contain "
                            + this.toString(name[0]));
                    NameComponent[] namesMinusObject = this.cropObject(name);
                    throw new NotFound(NotFoundReason.missing_node,
                                       namesMinusObject);
                }
            }
        } catch(NotFound e) {
            logger.debug("Context not found. Try creating it first.");
            throw e;
        }
    }

    private boolean containsName(NameComponent name) {
        return this.subcontext.containsKey(this.toString(name));
    }

    private NameComponent[] cropName(NameComponent[] name) {
        NameComponent[] newName = new NameComponent[name.length - 1];
        for(int i = newName.length; i > 0; i--) {
            newName[i - 1] = name[i];
        }
        return newName;
    }

    private NameComponent[] cropObject(NameComponent[] name) {
        NameComponent[] namesMinusObject = new NameComponent[name.length - 1];
        for(int i = 0; i < name.length - 1; i++) {
            namesMinusObject[i] = name[i];
        }
        return namesMinusObject;
    }

    public void destroy() {
        throw new NO_IMPLEMENT();
    }

    private String serverPropsName;

    private Properties confProps;

    private POA poa;

    private ORB orb;

    private HashMap subcontext, corbaObject;

    private static final Logger logger = LoggerFactory.getLogger(NamingContextImpl.class);
}
