package edu.sc.seis.cormorant.redirect;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import edu.iris.Fissures.IfNetwork.NetworkAccess;
import edu.iris.Fissures.IfNetwork.NetworkFinder;
import edu.iris.Fissures.IfNetwork.NetworkFinderPOA;
import edu.iris.Fissures.IfNetwork.NetworkId;
import edu.iris.Fissures.IfNetwork.NetworkNotFound;
import edu.iris.Fissures.network.NetworkIdUtil;
import edu.sc.seis.fissuresUtil.cache.ProxyNetworkAccess;
import edu.sc.seis.fissuresUtil.cache.VestingNetworkDC;
import edu.sc.seis.fissuresUtil.namingService.FissuresNamingService;

public class RedirectingNetworkFinder extends NetworkFinderPOA {

    public RedirectingNetworkFinder(Properties confProps, String controllerName) {
        this.props = confProps;
        this.controllerName = controllerName;
        logger = LoggerFactory.getLogger(RedirectingNetworkFinder.class.getName()
                + "." + controllerName);
    }

    public void start(FissuresNamingService nameService) throws NetworkNotFound {
        this.ns = nameService;
        String defaultServerName = props.getProperty(controllerName
                + ".base_server");
        logger.debug("loading base net server " + defaultServerName);
        NetworkFinder baseFinder = getFinder(defaultServerName);
        logger.debug("retrieving all from base");
        NetworkAccess[] baseNets = baseFinder.retrieve_all();
        List curNets = new ArrayList();
        for(int i = 0; i < baseNets.length; i++) {
            curNets.add(baseNets[i]);
        }
        String[] authServers = new String[0];
        if(props.containsKey(controllerName + ".authoritative_servers")) {
            authServers = props.getProperty(controllerName
                    + ".authoritative_servers").split(",");
        }
        for(int i = 0; i < authServers.length; i++) {
            logger.debug("loading authoritative net server " + authServers[i]);
            NetworkFinder authFinder = getFinder(authServers[i]);
            String[] codes = props.getProperty(controllerName + "."
                    + authServers[i] + ".codes").split(",");
            for(int j = 0; j < codes.length; j++) {
                logger.debug("setting " + authServers[i] + " for " + codes[j]);
                Iterator it = curNets.iterator();
                while(it.hasNext()) {
                    NetworkAccess cur = (NetworkAccess)it.next();
                    if(cur.get_attributes().get_code().equals(codes[j])) {
                        it.remove();
                    }
                }
                NetworkAccess[] authNets = authFinder.retrieve_by_code(codes[j]);
                for(int k = 0; k < authNets.length; k++) {
                    curNets.add(authNets[k]);
                }
            }
        }
        proxyNets = new NetworkAccess[curNets.size()];
        realNets = new NetworkAccess[curNets.size()];
        for(int i = 0; i < proxyNets.length; i++) {
            realNets[i] = ((ProxyNetworkAccess)curNets.get(i)).getCorbaObject();
            proxyNets[i] = (NetworkAccess)curNets.get(i);
        }
    }

    public void clear() {
        proxyNets = null;
        realNets = null;
    }

    private NetworkFinder getFinder(String serverName) {
        String name = props.getProperty(controllerName + "." + serverName
                + ".name");
        String dns = props.getProperty(controllerName + "." + serverName
                + ".dns");
        return new VestingNetworkDC(dns, name, ns).a_finder();
    }

    public NetworkAccess[] retrieve_all() {
        logger.debug("retrieve_all");
        return realNets;
    }

    public NetworkAccess[] retrieve_by_code(String requiredCode)
            throws NetworkNotFound {
        logger.debug("retrieve_by_code " + requiredCode);
        List toReturn = new ArrayList();
        for(int i = 0; i < proxyNets.length; i++) {
            if(proxyNets[i].get_attributes().get_code().equals(requiredCode)) {
                toReturn.add(realNets[i]);
            }
        }
        logger.debug("retrieve_by_code found " + toReturn.size()
                + " matching nets");
        return (NetworkAccess[])toReturn.toArray(new NetworkAccess[0]);
    }

    public NetworkAccess retrieve_by_id(NetworkId requiredId)
            throws NetworkNotFound {
        logger.debug("retrieve_by_id " + NetworkIdUtil.toString(requiredId));
        for(int i = 0; i < proxyNets.length; i++) {
            if(NetworkIdUtil.areEqual(requiredId, proxyNets[i].get_attributes()
                    .get_id())) {
                return realNets[i];
            }
        }
        throw new NetworkNotFound("No network of id "
                + NetworkIdUtil.toString(requiredId));
    }

    public NetworkAccess[] retrieve_by_name(String requiredName) {
        logger.debug("retrieve_by_name " + requiredName);
        List toReturn = new ArrayList();
        for(int i = 0; i < proxyNets.length; i++) {
            if(proxyNets[i].get_attributes().getName().equals(requiredName)) {
                toReturn.add(realNets[i]);
            }
        }
        logger.debug("retrieve_by_name found " + toReturn.size()
                + " matching nets");
        return (NetworkAccess[])toReturn.toArray(new NetworkAccess[0]);
    }

    private NetworkAccess[] proxyNets, realNets;

    private Properties props;

    private String controllerName;

    private FissuresNamingService ns;

    private final Logger logger;
}
