package edu.sc.seis.cormorant.seismogram;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import org.omg.CORBA.NO_IMPLEMENT;

import edu.iris.Fissures.FissuresException;
import edu.iris.Fissures.Time;
import edu.iris.Fissures.IfNetwork.ChannelId;
import edu.iris.Fissures.IfSeismogramDC.DataCenterCallBack;
import edu.iris.Fissures.IfSeismogramDC.DataCenterOperations;
import edu.iris.Fissures.IfSeismogramDC.LocalSeismogram;
import edu.iris.Fissures.IfSeismogramDC.RequestFilter;
import edu.iris.Fissures.model.MicroSecondDate;
import edu.iris.Fissures.model.QuantityImpl;
import edu.iris.Fissures.model.SamplingImpl;
import edu.iris.Fissures.model.TimeInterval;
import edu.iris.Fissures.model.UnitImpl;
import edu.iris.Fissures.network.ChannelIdUtil;
import edu.iris.Fissures.seismogramDC.LocalSeismogramImpl;
import edu.iris.Fissures.seismogramDC.RequestFilterUtil;
import edu.sc.seis.fissuresUtil.cache.WorkerThreadPool;
import edu.sc.seis.fissuresUtil.chooser.ClockUtil;
import edu.sc.seis.fissuresUtil.comparator.SeisBeginTimeComparator;
import edu.sc.seis.fissuresUtil.display.MicroSecondTimeRange;
import edu.sc.seis.fissuresUtil.exceptionHandler.GlobalExceptionHandler;
import edu.sc.seis.fissuresUtil.mseed.FissuresConvert;
import edu.sc.seis.fissuresUtil.time.CoverageTool;
import edu.sc.seis.fissuresUtil.time.ReduceTool;
import edu.sc.seis.seisFile.mseed.Blockette100;
import edu.sc.seis.seisFile.mseed.Blockette1000;
import edu.sc.seis.seisFile.mseed.Blockette2000;
import edu.sc.seis.seisFile.mseed.DataHeader;
import edu.sc.seis.seisFile.mseed.DataRecord;
import edu.sc.seis.seisFile.mseed.SeedFormatException;
import edu.sc.seis.seisFile.mseed.Utility;

public class BudLightCachingDC extends BudLightDataCenter {

    public BudLightCachingDC(BudLightWriter writer, DataCenterOperations remoteOps) {
        super(writer.getDataDir());
        this.remoteOps = remoteOps;
        this.queue = writer.getQueue();
    }
    
    

    public RequestFilter[] available_data(RequestFilter[] request) {
        logger.debug("available_data "+request.length);
        List<RequestFilter> out = new ArrayList<RequestFilter>();
        Map<String, List<RequestFilter>> splitRF = RequestFilterUtil.splitByChannel(Arrays.asList(request));
        for (String key : splitRF.keySet()) {
            out.addAll(availableDataOneChannel(splitRF.get(key)));
        }
        RequestFilter[] seis =  out.toArray(new RequestFilter[0]);
        seis = ReduceTool.merge(seis);
        logger.debug("Done available_data aaa"+request.length);
        return seis;
    }

    public List<RequestFilter> availableDataOneChannel(List<RequestFilter> request) {
       /* this is bad, but TMDB is really screwed up in the BUD, so just send back "no data" */
        if (request.get(0).channel_id.station_code.equals("TMDB")) {
            logger.warn("Ignoring request for TMDB");
            return new ArrayList<RequestFilter>();
        }

        if ( RequestFilterUtil.containsWildcard(request)) {
            // don't look locally for wildcards as we have to go remote anyway as are not sure of complete channel list
            return Arrays.asList(remoteOps.available_data(request.toArray(new RequestFilter[0])));
        }
        List<DataRecord> noData = retrieveDataRecords(request, "nodata");
        RequestFilter[] knownNoData = new RequestFilter[noData.size()];
        for (int i = 0; i < knownNoData.length; i++) {
            knownNoData[i] = new RequestFilter(request.get(0).channel_id,
                                               FissuresConvert.getMicroSecondTime(noData.get(i).getHeader().getStartBtime()).getFissuresTime(),
                                               FissuresConvert.getMicroSecondTime(noData.get(i).getHeader().getLastSampleBtime()).getFissuresTime());
        }
        RequestFilter[] stillToCheck = CoverageTool.notCovered(request.toArray(new RequestFilter[0]), knownNoData);
        ArrayList<RequestFilter> out = new ArrayList<RequestFilter>();
        ArrayList<RequestFilter> localUnknown = new ArrayList<RequestFilter>();
        for (RequestFilter rf : stillToCheck) {
            RequestFilter[] local = super.available_data(new RequestFilter[] {rf});
            if (local.length != 0) {
                // this is bad as we assume anything means we know everything locally
                for (int j = 0; j < local.length; j++) {
                    out.add(local[j]);
                }
            } else {
                localUnknown.add(rf);
            }
        }
        if (availableDataGoRemote) {
            RequestFilter[] remote = remoteOps.available_data(localUnknown.toArray(new RequestFilter[0]));
            for (int j = 0; j < remote.length; j++) {
                out.add(remote[j]);
            }
            RequestFilter[] notCovered = CoverageTool.notCovered(request.toArray(new RequestFilter[0]),
                                                                 out.toArray(new RequestFilter[0]));
            for (int k = 0; k < notCovered.length; k++) {
                sendNoDataToCache(notCovered[k].channel_id, new MicroSecondTimeRange(notCovered[k]));
            }
        } else {
            // might be data remote, so we say yes
            out.addAll(localUnknown);
        }
        return out;
    }

    public LocalSeismogram[] retrieve_seismograms(RequestFilter[] request) throws FissuresException {
        logger.debug("retrieve_seismograms "+request.length);
        if (RequestFilterUtil.containsWildcard(Arrays.asList(request))) {
            throw new FissuresException(new edu.iris.Fissures.Error(1, "Wildcards not supported"));
        }
        List<LocalSeismogramImpl> out = new ArrayList<LocalSeismogramImpl>();
        Map<String, List<RequestFilter>> splitRF = RequestFilterUtil.splitByChannel(Arrays.asList(request));
        LocalSeismogramImpl[] seis;
        try {
        for (String key : splitRF.keySet()) {
            out.addAll(processOneChannel(splitRF.get(key)));
        }
        seis =  out.toArray(new LocalSeismogramImpl[0]);
        seis = ReduceTool.merge(seis);
        logger.debug("Done retrieve_seismograms aaa"+request.length);
        return seis;
        } catch (OutOfMemoryError e) {
           out = null;
           seis = null;
           logger.error("OutOfMemory", e);
           System.exit(11);
           // never happens
           throw e;
        }
    }
    

    public List<LocalSeismogramImpl> processOneChannel(List<RequestFilter> request) throws FissuresException {
       /* this is bad, but TMDB is really screwed up in the BUD, so just send back "no data" */
        if (request.get(0).channel_id.station_code.equals("TMDB")) {
            logger.warn("Ignoring request for TMDB");
            return new ArrayList<LocalSeismogramImpl>();
        }
        List<LocalSeismogramImpl> out = super.processOneChannel( request);
        List<RequestFilter> knownNoData = new ArrayList<RequestFilter>();
        logger.debug("Try for nodata files: ");
        List<DataRecord> nodataFromFile = retrieveDataRecords(request, "nodata");
        Iterator<DataRecord> it = nodataFromFile.iterator();
        while (it.hasNext()) {
            DataRecord dr = it.next();
            if (dr.getHeader().getDataQualityFlags() == MISSING_PADDED_DATA) {
                it.remove();
                knownNoData.add(new RequestFilter(request.get(0).channel_id,
                                                  new Time(FissuresConvert.getISOTime(dr.getHeader().getStartBtime()),
                                                           -1), new Time(FissuresConvert.getISOTime(dr.getHeader()
                                                          .getPredictedNextStartBtime()), -1)));
            }
        }
        RequestFilter[] notCovered = CoverageTool.notCovered(request.toArray(new RequestFilter[0]),
                                                             out.toArray(new LocalSeismogramImpl[0]));
        /*
            logger.info("REQUEST: "+RequestFilterUtil.toString(request));
            for (LocalSeismogramImpl seis : out) {
                logger.info("LOCAL SEIS: "+seis.getBeginTime()+" "+seis.getEndTime());
            }
            logger.info("KNOWN UNKOWNS: "+RequestFilterUtil.toString(knownNoData));
        */
        // remove "known unknowns" from notCovered
        notCovered = RequestFilterUtil.removeSmallRequests(CoverageTool.notCovered(notCovered, knownNoData.toArray(new RequestFilter[0])), MIN_REQUEST_SIZE);
        if (notCovered.length != 0) {  
            //logger.info("NOT COVERED: "+RequestFilterUtil.toString(notCovered));
            boolean moreThanMinRequest = false;
            for (int j = 0; j < notCovered.length; j++) {
                MicroSecondTimeRange timeRange = new MicroSecondTimeRange(notCovered[j]);
                if (timeRange.getInterval().greaterThan(maxGapInterval)) {
                    moreThanMinRequest = true;
                } else {
                    
                }
            }
            if (moreThanMinRequest && notCovered.length != 0) {  
                logger.debug("try remote for "+ChannelIdUtil.toStringNoDates(notCovered[0].channel_id)+" ask " + notCovered.length+" requests. "+RequestFilterUtil.toString(notCovered));
                try {
                    LocalSeismogram[] remote = remoteOps.retrieve_seismograms(notCovered);
                    logger.debug("try remote for "+ChannelIdUtil.toStringNoDates(notCovered[0].channel_id)+" got " + remote.length+" seismos.");
                    for (int k = 0; k < remote.length; k++) {
                        out.add((LocalSeismogramImpl)remote[k]);
                    }
                    sendToCache(remote);
                    RequestFilter[] stillNotCovered = CoverageTool.notCovered(notCovered, LocalSeismogramImpl.implize(remote));
                    for (int j = 0; j < stillNotCovered.length; j++) {
                        MicroSecondTimeRange timeRange = new MicroSecondTimeRange(stillNotCovered[j]);
                        if (timeRange.getEndTime().before(ClockUtil.now().subtract(maxDataLag))) {
                            RequestFilter[] notCoveredRemote = CoverageTool.notCovered(request.toArray(new RequestFilter[0]),
                                                                                       out.toArray(new LocalSeismogramImpl[0]));
                            for (int k = 0; k < notCoveredRemote.length; k++) {
                                sendNoDataToCache(notCoveredRemote[k].channel_id, new MicroSecondTimeRange(notCoveredRemote[k]));
                            }
                        } else {
                            logger.debug("No data, but still within max_lag, so not writing nodata marker");
                        }
                    }
                } catch (org.omg.CORBA.SystemException e) {
                    logger.warn("problem with remote for "+ChannelIdUtil.toStringNoDates(notCovered[0].channel_id)+" ask " + notCovered.length+" requests.");
                    throw new RuntimeException("Upstream server threw "+e.getClass()+" "+e.getMessage(), e);
                }
            }
        }
        return out;
    }

    public String request_seismograms(RequestFilter[] a_filterseq,
                                      DataCenterCallBack a_client,
                                      boolean long_lived,
                                      Time expiration_time) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public String queue_seismograms(RequestFilter[] a_filterseq) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public LocalSeismogram[] retrieve_queue(String a_request) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public void cancel_request(String a_request) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public String request_status(String a_request) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    protected void sendToCache(final LocalSeismogram[] seis) {
        pool.invokeLater( new Runnable() {
            public void run() {
                List<LocalSeismogramImpl> seisList = new ArrayList<LocalSeismogramImpl>();
                for (int i = 0; i < seis.length; i++) {
                    seisList.add((LocalSeismogramImpl)seis[i]);
                }
                Collections.sort(seisList, new SeisBeginTimeComparator());
                for (LocalSeismogramImpl lsimpl : seisList) {
                    try {
                        DataRecord[] dr = FissuresConvert.toMSeed(lsimpl);
                        for (int j = 0; j < dr.length; j++) {
                            queue.addWork(dr[j]);
                        }
                    } catch(SeedFormatException e) {
                        // oops
                        logger.error("Unable to convert to MiniSeed, skipping: "
                                     + ChannelIdUtil.toString(lsimpl.getChannelID()) + " from "
                                     + lsimpl.getBeginTime() + " to " + lsimpl.getEndTime(),
                                     e);
                    }
                }
            }
        });
    }
    
    protected void sendNoDataToCache(final ChannelId channel_id, MicroSecondTimeRange range) {
        logger.debug("no data: "+ChannelIdUtil.toStringNoDates(channel_id)+" "+range);
        MicroSecondDate begin = range.getBeginTime();
        Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        cal.setTime(range.getBeginTime());
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        cal.add(Calendar.HOUR, 1);
        MicroSecondDate end = new MicroSecondDate(cal.getTime());
        while(end.before(range.getEndTime())) {
            sendOneNoDataToCache(channel_id, new MicroSecondTimeRange(begin, end));
            begin = end.add(TINIEST_GAP);
            cal.add(Calendar.HOUR, 1);
            end = new MicroSecondDate(cal.getTime());
        }
        sendOneNoDataToCache(channel_id, new MicroSecondTimeRange(begin, range.getEndTime()));
    }
    
    protected void sendOneNoDataToCache(final ChannelId channel_id, MicroSecondTimeRange range) {   
        try {
            logger.debug("no data (hour break): "+ChannelIdUtil.toStringNoDates(channel_id)+" "+range);
            DataHeader header = new DataHeader(0, 'D', false);
            header.setDataQualityFlags(MISSING_PADDED_DATA);
            header.setStationIdentifier(channel_id.station_code);
            header.setLocationIdentifier(channel_id.site_code);
            header.setChannelIdentifier(channel_id.channel_code);
            header.setNetworkCode(channel_id.network_id.network_code);
            header.setStartBtime(FissuresConvert.getBtime(range.getBeginTime()));
            TimeInterval sampPeriod = range.getInterval();
            short[] multiAndFactor = FissuresConvert.calcSeedMultipilerFactor(new SamplingImpl(1, sampPeriod));
            header.setSampleRateFactor(multiAndFactor[0]);
            header.setSampleRateMultiplier(multiAndFactor[1]);
            Blockette1000 b1000 = new Blockette1000();
            b1000.setEncodingFormat((byte)3); // 3 is 32 bit int
            b1000.setWordOrder((byte)0);
            b1000.setDataRecordLength((byte)FissuresConvert.RECORD_SIZE_256_POWER);
            DataRecord dr = new DataRecord(header);
            dr.addBlockette(b1000);
            QuantityImpl hertz = sampPeriod.inverse().convertTo(UnitImpl.HERTZ);
            Blockette100 b100 = new Blockette100();
            b100.setActualSampleRate((float)hertz.getValue());
            dr.addBlockette(b100);
            Blockette2000 b2000 = new Blockette2000(new String[] {"no data exists: "+range}, new byte[0]);
            dr.addBlockette(b2000);
            header.setNumSamples((short)1);
            dr.setData(new byte[] {0x00, 0x00, 0x00, 0x00});
            String filename = BudLightWriter.calcFilename(dr.getHeader()).intern();
            synchronized(filename) {
                logger.debug("no data to "+filename);
                File f = new File(getDataDir(), filename);
                f.getParentFile().mkdirs();
                DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(f, true)));
                dr.write(dos);
                dos.close();
            }
        } catch(SeedFormatException e) {
            throw new RuntimeException("Should never happen.", e);
        } catch(FileNotFoundException e) {
            throw new RuntimeException("Should never happen.", e);
        } catch(IOException e) {
            throw new RuntimeException("Should never happen.", e);
        }
    }
    
    public static final byte MISSING_PADDED_DATA = (byte)16; // bit 4 means "Missing/padded data present"
    
    protected WorkerThreadPool pool = new WorkerThreadPool("seisToDR", 1); // only need one as writer is only one?

    DataRecordQueue queue;

    DataCenterOperations remoteOps;

 //   TimeInterval maxDataLag = new TimeInterval(2, UnitImpl.YEAR);
    TimeInterval maxDataLag = new TimeInterval(24, UnitImpl.HOUR);

    TimeInterval maxGapInterval = MAX_GAP;
    
    boolean availableDataGoRemote = false;
    
    static final int[] SINGLE_ZERO_ARRAY = new int[] { 0 };
    
    public static final TimeInterval MAX_GAP = new TimeInterval(5, UnitImpl.MINUTE);
    
    public static final TimeInterval MIN_REQUEST_SIZE = new TimeInterval(5, UnitImpl.MINUTE);
    
    public static final TimeInterval TINIEST_GAP = new TimeInterval(1, UnitImpl.TENTHMILLISECOND);
    
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(BudLightCachingDC.class);
    
    public TimeInterval getMaxDataLag() {
        return maxDataLag;
    }



    
    public void setMaxDataLag(TimeInterval maxDataLag) {
        this.maxDataLag = maxDataLag;
    }
}
