package edu.sc.seis.cormorant.seismogram;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import edu.sc.seis.fissuresUtil.comparator.DataRecordBeginComparator;
import edu.sc.seis.seisFile.mseed.DataHeader;
import edu.sc.seis.seisFile.mseed.DataRecord;

public class DataRecordQueue {

    Comparator<DataRecord> drComparator = new DataRecordBeginComparator();
    
    LinkedList<DataRecord> queue = new LinkedList<DataRecord>();

    int maxQueueSize = 100000;
    
    protected int queueSize = 0;
    
    public DataRecordQueue() {
        this(100000);
    }
    
    public DataRecordQueue(int maxSize) {
        maxQueueSize = maxSize;
    }

    public int getSize() {
        return queueSize;
    }
    
    public synchronized void addWork(DataRecord o) {
        while(getSize() > maxQueueSize) {
            try {
                notifyAll();
                wait();
            } catch(InterruptedException e) {
            }
        }
        addWorkNoWait(o);
    }

    public synchronized void addWorkNoWait(DataRecord o) {
        queue.addLast(o);
        queueSize++;
        notifyAll();
    }

    // Retrieve work from the work queue; block if the queue is empty
    public synchronized List<DataRecord> getWork() {
        while (queue.isEmpty()) {
            try {
                notifyAll();
                wait();
            } catch(InterruptedException e) {
            }
        }
        List<DataRecord> out = new ArrayList<DataRecord>();
        DataRecord dr = queue.removeFirst();
        queueSize--;
        out.add(dr);
        DataHeader drHead = dr.getHeader();
        Iterator<DataRecord> it = queue.iterator();
        while(it.hasNext()) { 
            DataRecord matchingDR = it.next();
            DataHeader matchingHeader = matchingDR.getHeader();
            if (drHead.getNetworkCode().equals(matchingHeader.getNetworkCode()) &&
                    drHead.getStationIdentifier().equals(matchingHeader.getStationIdentifier()) &&
                    drHead.getChannelIdentifier().equals(matchingHeader.getChannelIdentifier()) &&
                    drHead.getLocationIdentifier().equals(matchingHeader.getLocationIdentifier())) {
                it.remove();
                out.add(matchingDR);
                queueSize--;
            }
        }
        notifyAll();
        Collections.sort(out, drComparator);
        return out;
    }
    
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(DataRecordQueue.class);
}
